#include "ringbuffer.h"
#include "string.h"

#define MIN(a, b)    ((a) < (b) ? (a) : (b))

/** 
 * fls - find last (most-significant) bit set 
 * @x: the word to search 
 * 
 * This is defined the same way as ffs. 
 * Note fls(0) = 0, fls(1) = 1, fls(0x80000000) = 32. 
 */  
static inline int fls(int x)
{
    int r = 32;
    if (!x)
        return 0;
    if (!(x & 0xffff0000u)) {
        x <<= 16;
        r -= 16;        
    } 
    if (!(x & 0xff000000u)) {
        x <<= 8;
        r -= 8;
    }
    if (!(x & 0xf0000000u)) {
        x <<= 4;
        r -= 4;
    }
    if (!(x & 0xc0000000u)) {
        x <<= 2;
        r -= 2;
    }
    if (!(x & 0x80000000u)) {
        x <<= 1;
        r -= 1;
    }
    
    return r;
}

static unsigned int RoundOfTwo(unsigned int x)
{
#if use_dynamic_memory_alloc
    return (1UL << fls(x));
#else
    return (1UL << (fls(x) - 1));
#endif
}

int RingBufferInit(RingBuffer_t rb, char *buf, unsigned int buf_size)
{
    if (!rb || !buf || buf_size < 2) 
        return -1;

    rb->pool = buf;
    rb->read_index = 0;
    rb->write_index = 0;

    if (buf_size & (buf_size - 1)) {
        buf_size = RoundOfTwo(buf_size);
    }
    rb->size = buf_size;
    rb->mask_index = buf_size - 1;
    
    return 0;
}

int RingBufferWrite(RingBuffer_t rb, char *buf, int length)
{
    // 获取环形缓冲区剩余空间
    int writeLength = RingBufferSpace(rb);
    // 没有剩余空间直接返回
    if (!writeLength) 
        return -1;
    // 传入数据长度大于剩余空间时只存入剩余空间的大小
    length = MIN(writeLength, length);
    // 写入数据长度需要越过缓冲区最后一个地址时，分两次写入
    writeLength = MIN(length, rb->size - (rb->write_index & rb->mask_index));
    // 第一次只写入到缓冲区的最后一个地址
    memcpy(rb->pool+(rb->write_index & rb->mask_index), buf, writeLength);
    // 如果没有越过最后一个地址，length - writeLength为0，以下函数不执行
    memcpy(rb->pool, buf + writeLength, length - writeLength);

    rb->write_index += length;
    return 0;
}

int RingBufferReadByte(RingBuffer_t rb, char *buf, int length)
{
    int readLength = RingBufferDataLength(rb);
    if (!readLength) 
        return 0;
    
    length = MIN(readLength, length);
    readLength = MIN(length, rb->size - (rb->read_index & rb->mask_index));
    memcpy(buf, rb->pool + (rb->read_index & rb->mask_index), readLength);
    memcpy(buf + readLength, rb->pool, length - readLength);

    rb->read_index += length;
    
    return length;
}

int RingBufferDataLength(RingBuffer_t rb)
{
    return (rb->write_index - rb->read_index);
}

int RingBufferEmpty(RingBuffer_t rb)
{
    return (0 == RingBufferDataLength(rb));
}
