/*
 * This file is part of the Serial Flash Universal Driver Library.
 *
 * Copyright (c) 2016-2018, Armink, <armink.ztl@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * 'Software'), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 * IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 * CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 * TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * Function: Portable interface for each platform.
 * Created on: 2016-04-23
 */

#include <sfud.h>
#include <stdarg.h>

#include <stm32f1xx_hal.h>
#include "spi.h"
#include "gpio.h"
#include "string.h"
#include "stdio.h"

#define digitalLo(port, pin)  {port->BSRR = (uint32_t)pin << 16;}
#define digitalHi(port, pin)  {port->BSRR = pin;}

typedef struct{
	SPI_TypeDef *spix;
	GPIO_TypeDef *cs_gpiox;
	uint16_t cs_gpio_pin;
	SPI_HandleTypeDef *spi_handle;
}spi_user_data, *spi_user_data_t;

static char log_buf[256];

sfud_flash sfud_norflash0 = {
		.name = "norflash0",
		.spi.name = "SPI1",
		.chip = { "W25Q64CV", SFUD_MF_ID_WINBOND, 0x40, 0x17, 8L * 1024L * 1024L, SFUD_WM_PAGE_256B, 4096, 0x20},
};

void sfud_log_debug(const char *file, const long line, const char *format, ...);

static void gpio_configuration(spi_user_data_t spi)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	if(spi->spix == SPI1){
		GPIO_InitStructure.Pin = spi->cs_gpio_pin;
		GPIO_InitStructure.Mode = GPIO_MODE_OUTPUT_PP;
		GPIO_InitStructure.Pull = GPIO_PULLUP;
		GPIO_InitStructure.Speed = GPIO_SPEED_FREQ_HIGH;
		HAL_GPIO_Init(spi->cs_gpiox, &GPIO_InitStructure);
		HAL_GPIO_WritePin(spi->cs_gpiox, spi->cs_gpio_pin, GPIO_PIN_SET);
	}
}

static void spi_lock(const sfud_spi *spi)
{
	__disable_irq();
}

static void spi_unlock(const sfud_spi *spi)
{
	__enable_irq();
}

static void retry_delay_100us(void)
{
	uint32_t delay = 120;
	while(delay--);
}

static char log_buf[256];

void sfud_log_debug(const char *file, const long line, const char *format, ...);

/**
 * SPI write data then read data
 */
static sfud_err spi_write_read(const sfud_spi *spi, const uint8_t *write_buf, size_t write_size, uint8_t *read_buf,
        size_t read_size) {
    sfud_err result = SFUD_SUCCESS;
//    uint8_t send_data, read_data;

    /**
     * add your spi write and read code
     */
    HAL_StatusTypeDef state = HAL_OK;
    spi_user_data_t spi_dev = (spi_user_data_t)spi->user_data;

    /**
     * add your spi write and read code
     */

    if(write_size)
    {
    	SFUD_ASSERT(write_buf);
    }
    if(read_size)
    {
    	SFUD_ASSERT(read_buf);
    }

    digitalLo(spi_dev->cs_gpiox,spi_dev->cs_gpio_pin);

    if(write_size)
    {
    	state = HAL_SPI_Transmit(spi_dev->spi_handle, (uint8_t *)write_buf, write_size, 1000);
    	while(HAL_SPI_GetState(spi_dev->spi_handle) != HAL_SPI_STATE_READY);
    }

    if(state != HAL_OK)
    {
    	goto __exit;
    }

    if(read_size)
    {
    	memset((uint8_t *)read_buf, 0xFF, read_size);
    	state = HAL_SPI_Receive(spi_dev->spi_handle, read_buf, read_size, 1000);
    	while(HAL_SPI_GetState(spi_dev->spi_handle) != HAL_SPI_STATE_READY);
    }

__exit:

    digitalHi(spi_dev->cs_gpiox,spi_dev->cs_gpio_pin);

    return result;
}

#ifdef SFUD_USING_QSPI
/**
 * read flash data by QSPI
 */
static sfud_err qspi_read(const struct __sfud_spi *spi, uint32_t addr, sfud_qspi_read_cmd_format *qspi_read_cmd_format,
        uint8_t *read_buf, size_t read_size) {
    sfud_err result = SFUD_SUCCESS;

    /**
     * add your qspi read flash data code
     */

    return result;
}
#endif /* SFUD_USING_QSPI */

static spi_user_data spi1 = {.spix = SPI1, .cs_gpiox = GPIOA, .cs_gpio_pin = GPIO_PIN_4, .spi_handle = &hspi1};


sfud_err sfud_spi_port_init(sfud_flash *flash) {
    sfud_err result = SFUD_SUCCESS;

    /**
     * add your port spi bus and device object initialize code like this:
     * 1. rcc initialize
     * 2. gpio initialize
     * 3. spi device initialize
     * 4. flash->spi and flash->retry item initialize
     *    flash->spi.wr = spi_write_read; //Required
     *    flash->spi.qspi_read = qspi_read; //Required when QSPI mode enable
     *    flash->spi.lock = spi_lock;
     *    flash->spi.unlock = spi_unlock;
     *    flash->spi.user_data = &spix;
     *    flash->retry.delay = null;
     *    flash->retry.times = 10000; //Required
     */
    switch(flash->index){
    case SFUD_W25Q64_DEVICE_INDEX:
    	MX_SPI1_Init();
    	gpio_configuration(&spi1);

    	flash->spi.wr = spi_write_read;
    	flash->spi.lock = spi_lock;
    	flash->spi.unlock = spi_unlock;
    	flash->spi.user_data = &spi1;
    	flash->retry.delay = retry_delay_100us;
    	flash->retry.times = 60 * 10000;
    	break;
    }

    return result;
}

/**
 * This function is print debug info.
 *
 * @param file the file which has call this function
 * @param line the line number which has call this function
 * @param format output format
 * @param ... args
 */
void sfud_log_debug(const char *file, const long line, const char *format, ...) {
    va_list args;

    /* args point to the first variable parameter */
    va_start(args, format);
    printf("[SFUD](%s:%ld) ", file, line);
    /* must use vprintf to print */
    vsnprintf(log_buf, sizeof(log_buf), format, args);
    printf("%s\n", log_buf);
    va_end(args);
}

/**
 * This function is print routine info.
 *
 * @param format output format
 * @param ... args
 */
void sfud_log_info(const char *format, ...) {
    va_list args;

    /* args point to the first variable parameter */
    va_start(args, format);
    printf("[SFUD]");
    /* must use vprintf to print */
    vsnprintf(log_buf, sizeof(log_buf), format, args);
    printf("%s\n", log_buf);
    va_end(args);
}
